/*	Copyright: autocart (user name at the ahk and XYplorer forums) unless otherwise noted in the code
Version: 3.3
Date: 2020-04-28

General description:

Preamble:
Most of the code was written years ago. Looking at the code now, it does not look pretty at all. I feel a need to rewrite it from scratch but don't have the time nor the motivation for it all at the moment. For now, I just started a little bit and mostly changed the things, which seemed to me most important. In a later version I might rewrite the code - or not.

Purpose and usage of the script:
This script works in combination with the file explorer XYplorer (from now on called "XY") and Windows Explorer (from now on called "WE").
It was written for the following scenarios:

(1)	XY's setting "XY is default file manager" not getting respected:
	Despite XY's build-in optional functionality to act as the system's default file manager (XY menu Tools/Configuration/Other/Shell Integration), sometimes there still are WE windows opening instead of XY (e.g. from within certain 3rd-party-software or for other reasons).

(2)	List items not getting automatically selected in XY:
	At other times XY *is* sucessfully used as the system's default file manager to open a folder (if that setting is on) but any list itmes which would normally be selected in WE do *not* get selected in XY (e.g. when clicking in Firefox in the "downloads" window on the "open containing folder" button).

For such scenarios, this script, if running in the system tray, catches any newly opening WE window, reads the path and selected items shown in that WE window, closes that WE window, opens a new tab in XY with that path *and* selects any list items in XY that where selected in WE before (and scrolling them into view), thus "redirecting" the WE window to XY.

The script will stay in the system tray until manually exited through the script's tray context menu.

Hint:
For the script to work also in cases of scenario No. (2), of course, the XY setting to act as the system's default file manager must be turned *off*.

More details:

If one wants any WE windows to be kept as WE windows (and *not* be redirected to XY) - e.g. control panel stuff - then there are these options:

(1)	Control-key-escape:
	Hold the left or right control key pressed down while the WE window opens, which you want to keep. Typically, though, the control key must be held down until the WE window has actually opened. If released too early, it might not work as intended.

	Example:
	With the script running, pressing Win+E would open a new WE window which normally would get redirected to XY.
	Pressing Control+Win+E, the new WE window would *not* get redirected to XY. (Of course the Control-key-escape option should work also if a WE window is opened by other means.)

(2)	Pre-defined keywords / paths:
	The script allows for pre-defined keywords and paths, which are compared to the location being displayed in the WE window (typically the content of the WE address bar, e.g. "Computer", "Libraries", "Pictures" ... and paths without trailing backslash). These keywords and paths can be either set in the variable vWindowsExplorerPaths in the "USER CUSTOMIZE SECTION" of the code below - separated by the pipe character "|". Else they can also be passed to the script as command line parameters - separated by spaces (and quoted if they contain spaces). Both kinds of definition *are* allowed together at the same time.

	Example for the command line parameters option:
	In case of command line parameters (or parameters passed in the target line of a windows shortcut file) the command line could look like this:
	"path_to_autohotkey.exe" "path_to_script.ahk" Computer Libraries Pictures "C:\some kind of folder\another folder"

	Hint:
	In the "USER CUSTOMIZE SECTION" of the code below there are some suggestions for keywords to be used on an English version of Windows.

	The variable vKeepOrRedirect_matchingWEPaths in the "USER CUSTOMIZE SECTION" of the code below determines what will happen in case of a match (when comparing the WE window's location with the keywords/paths definitions). It must be set to either "Keep" or "Redirect". If vKeepOrRedirect_matchingWEPaths is set to "Keep", then the matching WE windows are kept as WE windows (and the others are redirected to XY). If vKeepOrRedirect_matchingWEPaths is set to "Redirect" then the matching WE windows are redirected to XY (and the others are kept as WE windows).
	
	If the user wishes, the method for comparing the WE window's location with the keywords/paths definitions, can be set to use Regular Expressions (RegEx). To do so, the variable vUseRegExForPathCompare in the "USER CUSTOMIZE SECTION" of the code below must be set to 1. Otherwise, if vUseRegExForPathCompare is set to 0, then - to be a match - a keyword or path and the WE window's location must be identical.
	
	Hint:
    Of course, if RegEx is used as a compare method, then also the keywords and paths must be defined using RegEx syntax.
	Here are some helpful sites for learning RegEx (PCRE flavor):
	https://www.regular-expressions.info/quickstart.html
	https://regex101.com/
	
	If RegEx is used as a compare method, then any additional keywords/paths defined in the extra variable vExceptionPathsInCaseOfRegEx in the "USER CUSTOMIZE SECTION" of the code below act as a subset of exceptions to the main keywords/paths definitions defined in vWindowsExplorerPaths. Therefore, the main definitions in vWindowsExplorerPaths can contain a more general pattern and the extra variable vExceptionPathsInCaseOfRegEx can contain exceptions from that general pattern. (This is *only* available for the compare method RegEx.)

Even more details:

WE windows, which show the content of a zip file, are automatically kept as WE windows if the "identical" compare method is used. This is, because XY does not support this kind of view and the "identical" compare method is not capable of defining zip files as a pattern. However, if RegEx is used as compare method, then the user must define it manually in the RegEx pattern if he/she wants a zip file-content-view to be kept in WE (e.g. with \.zip$ either as part of a Keep-pattern or as part of a exeptionPaths-when-Redirect-pattern).

If XY is not running it will be started automatically if the path to XYplorer.exe is specified correctly. The path can be specified in the variable vPathToXYplorerForAutostart in the "USER CUSTOMIZE SECTION" of the code below. A tooltip close to the mouse cursor will indicate the starting process of XY. If the XY-autostart should not work for some reason (e.g. if the wrong or no path is specified) the script will display a message box and wait for the user to start XY manually.

If the user wishes for the script to apply auto-redirection also to all WE windows already existing at the moment of the script's start (not just to newly created ones while it is running) then this can be activated in the variable vApplyRulesToExistingWEwindowsAtScriptStart in the "USER CUSTOMIZE SECTION" of the code below. Setting the variable to 1 will activate that behaviour. Setting it to 0 will deactivate it.

In order to manually redirect a given WE window to XY by force (regardless of the rules for automatic redirection), the user can specify a hotkey in the variable vHotkeyInWEwindowsToTriggerRedirectionToXY in the "USER CUSTOMIZE SECTION" of the code below. (A hotkey refers to pressing a combination of keys on the keyboard, namely a character key together with one or more modifier keys. Modifier keys are Control, Alt, Shift and the Windows key.) The WE window will be redirected if the hotkey is pressed while the WE window is active. More details on how to define the hotkey are given right next to the variable itself. This hotkey can be activated or deactivated on its own (just in case) in the script's system tray context menu. If no such hotkey is specified, then the system tray context menu item will not exist.

The whole script with all its functionality can be paused - respectively suspended - in the script's system tray context menu.

Known limitations:

-)	Whenever a WE window is "redirected" to XY, it will still open for a short moment in a normal way and immediately be closed again. For this script to work, I do not know of any way how to avoid that.
-)	In case of multiple XY windows being open at the same time, the most recently active XY window will be used for redirection.
-)	This script has not been tested with WE add-ons, which add tabs to WE.

For suggestions and ideas for improvement I am grateful.

Changes in version 3.0:

-)	Improved reliability:
	Before, the script sometimes missed WE windows or seemingly could not retrieve its path and then did not react as expected.
-)	Improved behaviour when paused:
	Now the script is respecting the script's paused state (see the script's system tray context menu) in that it stops redirecting when the script paused.
	Before it ignored the state of being paused and redirected anyway.
-)	Changed the behaviour of the script regarding already existing WE windows:
	In previous versions, when the title bar of already existing WE windows (which existed prior to the scripts start) changed (e.g. when changing a folder in such a window or when refreshing by hitting F5), then the script also used to redirect such WE windows to XY.
	Not anymore. Now, if a WE window already exist at the scripts start, the script leaves it untouched.
-)	Added a Control-key-escape option:
	If the left or right Control key is pressed while a WE window opens, then this particular WE window will *not* get redirected to XY.

Changes in version 3.1:

-)	Fixed the problem that the script would crash if XY was not active.
-)	The path to XY can now be specified in the variable $pathToXYplorer at the beginning of the code.
-)	In case of a XY autostart process, a tooltip will be diplayed close to the mouse cursor.
-)	In case the XY autostart process could not be started, a message box will pop.

Changes in version 3.2:

-)	Added code for an optional custom tray icon and supplied a simple icon with it.
-)	Renamed the variables; especially the variable with all the keywords and paths is now called vWindowsExplorerPaths.
-)	Added the options/variables vKeepOrRedirect_matchingWEPaths, vUseRegExForPathCompare and vExceptionPathsInCaseOfRegEx.
-)	Added some visual marking of the user customize section in the code below
-)	In case the variable vKeepOrRedirect_matchingWEPaths has an incorrect value, a message box will pop.
-)	Changed the hard coded behaviour of *not* redirecting the display of the content of zip files to XY: Now it is only hard coded for the case if RegEx is *not* used as compare method. If RegEx is used, then the behaviour for zip files can be defined in the RegEx pattern.

Changes in version 3.3:

-)  Fixed some wrong logic in the processing of events. Previously, if XY was not already running, it would be automatically started also if the WE window was to be kept as WE window. In that scenario also, if no correct path to XYplorer.exe was defined, then the script would still pop the corresponding message box and wait for XY to be started manually. This made not even any sense, since the WE window would anyway not be redirected to XY. This is changed now. The script will only check for the open XY window and react accordingly if the WE window is to be redirected.
-)  Fixed a logical bug with RegEx compare: If hard coded keywords/paths were used AND command line parameters AND RegEx was used as compare method, then there was a logical bug which could have led to everything being a match (besides the RegEx-exeptions)
-)  Changed the order of the help text a bit. Adjusted the order of the variables in the "USER CUSTOMIZE SECTION" of the code accordingly.
-)  Added the variable vApplyRulesToExistingWEwindowsAtScriptStart to optionally apply the defined rules also to existing WE windows at script start.
-)  Added the variable vHotkeyInWEwindowsToTriggerRedirectionToXY to be able to define a hotkey to redirect WE windows to XY manually, if pressed while the targeted WE window is active.
-)  Simplified the system tray context menu to three items:
	(1) Toggle menu item to use/activate the specified hotkey for redirecting WE windows manually (or deactivate it).
	(2) Toggle menu item to Pause / Suspend Script which both pauses the script and suspends the hotkeys.
	(3) Fully exit the script.

Possible ideas for the future:

-)  maybe handle also other message scenarios - not just "window created"
-)  hotkey(s) - either global or only when XY is active? - for pausing and un-pausing the script
-)  improve the MsgBox popping when the WE path could not be determined (maybe with debug-information that a user could then post in the forum for my knowledge) and make it easier for the end user to disable that MsgBox if he/she would want to not see it
-)  implement an option somehow to "move" XY paths in the other direction to Windows Explorer. If at all, then this would probably only make sense with a manual trigger, not automatically for pre-defined paths, or???
*/

;;;******************************************************************************************
;;;*********************   START of USER CUSTOMIZE SECTION - START  *************************
;;;******************************************************************************************

; vWindowsExplorerPaths := "[A-Z]:\\.*|This Computer"
vWindowsExplorerPaths := "|C:\sample path|All Control Panel Items|Storage Spaces|Sync Center|System|Network and Sharing Center|Programs and Features|Recovery|RemoteApp and Desktop Connections|Security and Maintenance|Speech Recognition|Autoplay|Fonts|Backup and Restore (Windows 7)|Devices and Printers|Dolby Audio|Ease of Access Center|File History|BitLocker Drive Encryption|Color Management|Credential Manager|Device Manager|Action Center|Administrative Tools|Default Programs|Mouse|Power Options|Personalization|User Accounts|Windows 10 Firewall Control|Troubleshooting|Windows Defender Firewall|Windows Firewall|Work Folders|"

vKeepOrRedirect_matchingWEPaths := "Keep"
; vKeepOrRedirect_matchingWEPaths := "Redirect"

vUseRegExForPathCompare := 0
vExceptionPathsInCaseOfRegEx := "\.zip$"

vPathToXYplorerForAutostart := "C:\Program Files (x86)\XYplorer\XYplorer.exe"

vApplyRulesToExistingWEwindowsAtScriptStart := 0

;; Hotkey-modifiers: "^" stands for Ctrl, "!" stands for Alt, "+" stands for Shift and "#" stands for the Windows key.
;; The hotkey itself *is* case sensitive! For a capital char it is better to use the small char together with the "+" modifier.
;; Extreme expamle: "^!+#h" would stand for Ctrl+Alt+Shift+Win+h (respectively Ctrl+Alt+Win+H)
vHotkeyInWEwindowsToTriggerRedirectionToXY := "^y"
; vHotkeyInWEwindowsToTriggerRedirectionToXY := ""

;;;******************************************************************************************
;;;**********************   END of USER CUSTOMIZE SECTION - END  ****************************
;;;******************************************************************************************



#NoEnv  ; Recommended for performance and compatibility with future AutoHotkey releases.
#SingleInstance force
#Persistent
; #Warn  ; Enable warnings to assist with detecting common errors.
SendMode Input  ; Recommended for new scripts due to its superior speed and reliability.
SetWorkingDir %A_ScriptDir%  ; Ensures a consistent starting directory.

vKeepIfCtrlHeldDown := 1

if (not (vKeepOrRedirect_matchingWEPaths = "Keep" or vKeepOrRedirect_matchingWEPaths = "Redirect"))
{
	MsgBox, 262144, %A_ScriptName%, The variable vKeepOrRedirect_matchingWEPaths must be either "Keep" or "Redirect", however it is "%vKeepOrRedirect_matchingWEPaths%".`n`nPlease, correct it in the user customize section of the code.`n`nThe script will terminate as soon as this dialog is closed.
	ExitApp
}

;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;; ~~~~~~~~~~~~~~~~~~~~~~   INITIALIZATION of TRAY MENU - START   ~~~~~~~~~~~~~~~~~~~~~~~~~~
;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
if !A_IsCompiled
{
	Menu, Tray, UseErrorLevel
	Menu, Tray, Icon, %A_ScriptDir%\redirectWEwindowsToXY.ico
}

Menu, Tray, NoStandard
if (vHotkeyInWEwindowsToTriggerRedirectionToXY)
{
	vDisplayFormatOfHKIWETTRTXY := StrReplace(StrReplace(StrReplace(StrReplace(vHotkeyInWEwindowsToTriggerRedirectionToXY, "#", "Win+"), "+", "Shift+"), "!", "Alt+"), "^", "Ctrl+")
	Menu, Tray, Add, Use %vDisplayFormatOfHKIWETTRTXY% to redirect WE to XY, _ToggleHotkeyIWETTRTXY
	Menu, Tray, Check, Use %vDisplayFormatOfHKIWETTRTXY% to redirect WE to XY
}
; Menu, Tray, Add, Suspend Hotkeys, _ToggleSuspendHotkeys
Menu, Tray, Add, Pause / Suspend Script, _TogglePauseAndSuspendScript
Menu, Tray, Add, Exit, _ExitApp
; Menu, Tray, Standard
;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;; ~~~~~~~~~~~~~~~~~~~~~~   INITIALIZATION of TRAY MENU - END   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~


;; Before adding the command line parameters, ensure that the first and last char of vWindowsExplorerPaths is a pipe "|".
;; However, the end must be checked first if it is already a pipe, because if two consecutive pipes end up in the middle of the string (after command line parameters are added) then this would lead to unwanted results in case of the RegEx compare method. (compare below)
if (not SubStr(vWindowsExplorerPaths, 0) = "|")
	vWindowsExplorerPaths := "|" . vWindowsExplorerPaths . "|"
else
	vWindowsExplorerPaths := "|" . vWindowsExplorerPaths

/*	About incorrectly passed command line parameters:
	The help file says:
	Any parameter that contains spaces should be enclosed in quotation marks. A literal quotation mark may be passed in by preceding it with a backslash (\"). Consequently, any trailing slash in a quoted parameter (such as "C:\My Documents\") is treated as a literal quotation mark (that is, the script would receive the string C:\My Documents"). To remove such quotes, use StringReplace, 1, 1, ",, All.
	-> In my opinion that is a two edged sword. The core problem is that literal quotation marks are escaped with a backslash. Thus, two consecutive parameters of which the first ends on \" (backslash quote) are treated as one, no matter what. So, I comment the "correction" out in the next line for now.
	StringReplace, vWindowsExplorerPaths, vWindowsExplorerPaths, ",, All
	If wanted/needed, the whole raw command line can be correctly read using full_command_line := DllCall("GetCommandLine", "Str")
	Then it could be custom-parsed.
	For now, I go with instructing the user to NOT pass paths with a trailing backslash.
*/

For n, param in A_Args
{
    vWindowsExplorerPaths .= param . "|"
}

;; if RegEx is used as comparision method, then the pipes "|" at the beginning and end must be removed again, because otherwise the ends would be interpreted as empty patterns which would always match
if vUseRegExForPathCompare
	vWindowsExplorerPaths := Trim(vWindowsExplorerPaths, "|")


;; Make the GUI window the last found window for use by the line below.
Gui +LastFound
vHWnd := WinExist()
DllCall( "RegisterShellHookWindow", UInt, vHWnd )
vMsgNum := DllCall( "RegisterWindowMessage", Str, "SHELLHOOK" )
OnMessage( vMsgNum, "ShellMessage" )


if (vApplyRulesToExistingWEwindowsAtScriptStart)
{
	WinGet, vExistingCabinetWClassWindows, List, ahk_class CabinetWClass
	Loop %vExistingCabinetWClassWindows%
	{
		ShellMessage(1, vExistingCabinetWClassWindows%A_Index%)
	}
}


if (vHotkeyInWEwindowsToTriggerRedirectionToXY)
{
	Hotkey, IfWinActive, ahk_class CabinetWClass
	Hotkey, %vHotkeyInWEwindowsToTriggerRedirectionToXY%, _TriggerWEwindowRedirectionToXY
	Hotkey, If
}
return

;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;; ~~~~~~~~~~~~~~~~~~~~~~~~   END of AUTOEXECUTE SECTION   ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
;;; ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~



_TriggerWEwindowRedirectionToXY()
{
	global
	Critical
	if (not vHWndWE := WinActive("ahk_class CabinetWClass"))
		return
	vWindowsExplorerPaths_org := vWindowsExplorerPaths
	vWindowsExplorerPaths := ""
	vKeepOrRedirect_matchingWEPaths_org := vKeepOrRedirect_matchingWEPaths
	vKeepOrRedirect_matchingWEPaths := "Keep"
	vUseRegExForPathCompare_org := vUseRegExForPathCompare
	vUseRegExForPathCompare := 0
	vKeepIfCtrlHeldDown_org := vKeepIfCtrlHeldDown
	vKeepIfCtrlHeldDown := 0
	ShellMessage( 1, vHWndWE )
	vWindowsExplorerPaths := vWindowsExplorerPaths_org
	vKeepOrRedirect_matchingWEPaths := vKeepOrRedirect_matchingWEPaths_org
	vUseRegExForPathCompare := vUseRegExForPathCompare_org
	vKeepIfCtrlHeldDown := vKeepIfCtrlHeldDown_org
}

^#e::
;; Using this construct so that the user in this particular case does not *have* to keep pressing the Ctrl key down until the WE window appears.
OnMessage( vMsgNum, "ShellMessage", 0 )
OnMessage( vMsgNum, "ShellMessage_forCtrlWinE" )
Send, #e
return

ShellMessage_forCtrlWinE( wParam, lParam )
{
	Critical
	global vMsgNum
	;; Do nothing, thus the WE window is kept.
	;; Then (after doing nothing) set OnMessage back to the normal function.
	if (wParam = 1 and _WinGetClass("ahk_id " lParam) = "CabinetWClass")
	{
		OnMessage( vMsgNum, "ShellMessage_forCtrlWinE", 0 )
		OnMessage( vMsgNum, "ShellMessage" )
	}
}

ShellMessage( wParam, lParam )
{
	global
	Critical

/*wParam =
HSHELL_ACCESSIBILITYSTATE
11
The accessibility state has changed.

HSHELL_ACTIVATESHELLWINDOW
3
The shell should activate its main window.

HSHELL_APPCOMMAND
12
The user completed an input event (for example, pressed an application command button on the mouse or an application command key on the keyboard), and the application did not handle the WM_APPCOMMAND message generated by that input.
If the Shell procedure handles the WM_COMMAND message, it should not call CallNextHookEx. See the Return Value section for more information.

HSHELL_GETMINRECT
5
A window is being minimized or maximized. The system needs the coordinates of the minimized rectangle for the window.

HSHELL_LANGUAGE
8
Keyboard language was changed or a new keyboard layout was loaded.

HSHELL_REDRAW
6
The title of a window in the task bar has been redrawn.

HSHELL_TASKMAN
7
The user has selected the task list. A shell application that provides a task list should return TRUE to prevent Windows from starting its task list.

HSHELL_WINDOWACTIVATED
4
The activation has changed to a different top-level, unowned window.

HSHELL_WINDOWCREATED
1
A top-level, unowned window has been created. The window exists when the system calls this hook.

HSHELL_WINDOWDESTROYED
2
A top-level, unowned window is about to be destroyed. The window still exists when the system calls this hook.

HSHELL_WINDOWREPLACED
13
A top-level window is being replaced. The window exists when the system calls this hook. 

lParam differs in type according to the value of wParam received. For most of the wParam values, the lParam is a handle to a window that can be used as ahk_id %lParam% in AHK's Window commands.
*/

	; static vHWndsToIgnore := "|"

	; if ( (wParam = 1 or wParam = 6) and WinActive("ahk_class CabinetWClass"))
	if (wParam = 1 and _WinGetClass("ahk_id " lParam) = "CabinetWClass")
	{
		; if (InStr(vHWndsToIgnore, "|" lParam "|"))
		; 	return
		if (vKeepIfCtrlHeldDown and GetKeyState("Control", "P"))
		{
			; vHWndsToIgnore .= lParam "|"
			return
		}

		vPath := Explorer_GetPath(lParam)
		while ( vPath = "ERROR" and A_Index < 10)
		{
			 Sleep 100
			 vPath := Explorer_GetPath(lParam)
		}
		if ( vPath = "ERROR" )
		{
			 MsgBox Could not retrieve the path of the explorer window.
			 return
		}
		vSelected := Explorer_GetSelected(lParam)
		StringReplace, vSelected, vSelected, % Chr(10), |, 1
		StringReplace, vSelected, vSelected, ', '', 1


		;; If RegEx is *not* used as compare method, then keep WE in case of the content of a zip file
		if (not vUseRegExForPathCompare)
		{
			if (StrLen(vPath) > 3 and StrLen(vPath) - InStr(vPath, ".zip", 0, 0) = 3)
				return
		}

		
		if (vUseRegExForPathCompare)
		{
			; MsgBox, % vPath ", " vWindowsExplorerPaths ", " vKeepOrRedirect_matchingWEPaths
			if (RegExMatch(vPath, vWindowsExplorerPaths))
			{
				if (vKeepOrRedirect_matchingWEPaths = "Keep" and (not vExceptionPathsInCaseOfRegEx or not RegExMatch(vPath, vExceptionPathsInCaseOfRegEx)))
				{
					; MsgBox, 1
					return
				}
				else if (vKeepOrRedirect_matchingWEPaths = "Redirect" and vExceptionPathsInCaseOfRegEx and RegExMatch(vPath, vExceptionPathsInCaseOfRegEx))
				{
					; MsgBox, 2
					return
				}
				; MsgBox, 3
			}
			else if (vKeepOrRedirect_matchingWEPaths = "Redirect")
					return
		}
		else if (InStr(vWindowsExplorerPaths, "|" . vPath . "|"))
		{
			if (vKeepOrRedirect_matchingWEPaths = "Keep")
				return
		}
		else if (vKeepOrRedirect_matchingWEPaths = "Redirect")
			return

			
		SetTitleMatchMode, 2
		IfWinNotExist, XYplorer ahk_class ThunderRT6FormDC
		{
			SplitPath, vPathToXYplorerForAutostart, OutFileName, OutDir, OutExtension, OutNameNoExt, OutDrive
			; Run, XYPlorer.exe, OutDir, UseErrorLevel
			Run, %vPathToXYplorerForAutostart%, OutDir, UseErrorLevel
			if (ErrorLevel = "ERROR")
			{
				while not WinExist("XYplorer ahk_class ThunderRT6FormDC")
				{
					MsgBox, 262145, %A_ScriptName%, XYplorer could not be started automatically. Please, start it manually.`n`nOnce XYplorer is started you can close this dialog in order to redirect the Windows Explorer window to XYplorer.`n`nPress Cancel in order to *not* redirect the Windows Explorer window to XYplorer this time.
					IfMsgBox, Cancel
						return
				}
			}
			else
			{
				; WinWait, XYplorer ahk_class ThunderRT6FormDC
				while not WinExist("XYplorer ahk_class ThunderRT6FormDC")
				{
					ToolTip, %A_ScriptName%:`n`nXYplorer is being started...
					Sleep, 20
				}
				ToolTip
			}
			IfWinNotExist, ahk_id %lParam%
				return			
		}

		; WinWait, XYplorer ahk_class ThunderRT6FormDC
		WinActivate, XYplorer ahk_class ThunderRT6FormDC
		WinClose, ahk_id %lParam%
		XY_send_message_noreturnvalue("::if(strpos(get('tabs'), '" . vPath "') != -1){goto '" . vPath "', 1;}else{tab('new', '" . vPath . "');}")
		; XY_send_message_noreturnvalue("::goto '" . vPath "', 1;")
		XY_send_message_noreturnvalue("::selectitems '" . vSelected . "';#1012;") ;#1012 is XY-command-ID for Miscellaneous | Focus Functions | Move Focused Item Into View
	}
}


_WinGetClass(WinTitle := "", WinText := "", ExcludeTitle := "", ExcludeText := "")
{
	WinGetClass, class, %WinTitle%, %WinText%, %ExcludeTitle%, %ExcludeText%
	return class
}


_ToggleHotkeyIWETTRTXY()
{
	global
	Menu, Tray, ToggleCheck, Use %vDisplayFormatOfHKIWETTRTXY% to redirect WE to XY
	Hotkey, IfWinActive, ahk_class CabinetWClass
	Hotkey, %vHotkeyInWEwindowsToTriggerRedirectionToXY%, Toggle
	Hotkey, If
}

; _ToggleSuspendHotkeys()
; {
; 	global
; 	Menu, Tray, ToggleCheck, Suspend Hotkeys
; 	Suspend, Toggle
; }

_TogglePauseAndSuspendScript()
{
	global
	if A_IsPaused
		OnMessage( vMsgNum, "ShellMessage")
	else
		OnMessage( vMsgNum, "ShellMessage", 0 )
	Menu, Tray, ToggleCheck, Pause / Suspend Script
	Suspend, Toggle
	Pause, Toggle
}

_ExitApp()
{
	ExitApp
}


;*****************************************************************************************************************
;~ Code from others  (or based on):
;*****************************************************************************************************************
/*
	Library for getting info from a specific explorer window (if window handle not specified, the currently active
	window will be used).  Requires AHK_L or similar.  Works with the desktop.  Does not currently work with save
	dialogs and such.
	
	
	Explorer_GetSelected(hwnd="")   - paths of target window's selected items
	Explorer_GetAll(hwnd="")        - paths of all items in the target window's folder
	Explorer_GetPath(hwnd="")       - path of target window's folder
	
	example:
		F1::
			path := Explorer_GetPath()
			all := Explorer_GetAll()
			sel := Explorer_GetSelected()
			MsgBox % path
			MsgBox % all
			MsgBox % sel
		return
	
	Joshua A. Kinnison
	2011-04-27, 16:12
*/

Explorer_GetPath(hwnd="")
{
	if !(window := Explorer_GetWindow(hwnd))
		return ErrorLevel := "ERROR"
	if (window="desktop")
		return A_Desktop
	path := window.LocationURL
	if (path="")
		path:=window.LocationName
	path := RegExReplace(path, "ftp://.*@","ftp://")
	StringReplace, path, path, file:///
	StringReplace, path, path, /, \, All 
	
	; thanks to polyethene
	Loop
		if RegExMatch(path, "i)(?<=%)[\da-f]{1,2}", hex)
			StringReplace, path, path, `%%hex%, % Chr("0x" . hex), All
		Else Break
	return path
}

Explorer_GetAll(hwnd="")
{
	return Explorer_Get(hwnd)
}

Explorer_GetSelected(hwnd="")
{
	return Explorer_Get(hwnd,true)
}

Explorer_GetWindow(hwnd="")
{
	; thanks to jethrow for some pointers here
    WinGet, process, processName, % "ahk_id " hwnd := hwnd? hwnd:WinExist("A")
    WinGetClass class, ahk_id %hwnd%


	; MsgBox, %hwnd%, %process%, %class%
	; ; if not process
	; 	MsgBox, % WinExist("ahk_id " hwnd)
	; ; return
	
	if (process!="explorer.exe")
		return
	if (class ~= "(Cabinet|Explore)WClass")
	{
		for window in ComObjCreate("Shell.Application").Windows
			try if (window.hwnd==hwnd)
				return window
	}
	else if (class ~= "Progman|WorkerW") 
		return "desktop" ; desktop found
}

Explorer_Get(hwnd="",selection=false)
{
	if !(window := Explorer_GetWindow(hwnd))
		return ErrorLevel := "ERROR"
	if (window="desktop")
	{
		ControlGet, hwWindow, HWND,, SysListView321, ahk_class Progman
		if !hwWindow ; #D mode
			ControlGet, hwWindow, HWND,, SysListView321, A
		ControlGet, files, List, % ( selection ? "Selected":"") "Col1",,ahk_id %hwWindow%
		base := SubStr(A_Desktop,0,1)=="\" ? SubStr(A_Desktop,1,-1) : A_Desktop
		Loop, Parse, files, `n, `r
		{
			path := base "\" A_LoopField
			IfExist %path% ; ignore special icons like Computer (at least for now)
				ret .= path "`n"
		}
	}
	else
	{
		if selection
			collection := window.document.SelectedItems
		else
			collection := window.document.Folder.Items
		for item in collection
			ret .= item.path "`n"
	}
	return Trim(ret,"`n")
}

;*****************************************************************************************************************

;heavily based on
;http://www.xyplorer.com/xyfc/viewtopic.php?f=7&t=9233&p=89555&hilit=messenger#p89555

XY_send_message_noreturnvalue(MessagetoXYplorer)          ;Send message to XYplorer
{
  SenderHWND := A_ScriptHwnd + 0  ;Return this script's hidden hwdn id.  +0 to convert from Hex to Dec
  SetTitleMatchMode, 2
  HWND := WinExist("XYplorer ahk_class ThunderRT6FormDC")
  Size := StrLen(MessagetoXYplorer)
  if !(A_IsUnicode)
  {
  VarSetCapacity(Data, Size * 2, 0)
  StrPut(MessagetoXYplorer, &Data, Size, "UTF-16")
  }
  Else
  Data := MessagetoXYplorer

  VarSetCapacity(COPYDATA, A_PtrSize * 3, 0)
  NumPut(4194305, COPYDATA, 0, "Ptr")
  NumPut(Size * 2, COPYDATA, A_PtrSize, "UInt")
  NumPut(&Data, COPYDATA, A_PtrSize * 2, "Ptr")
  SendMessage, 0x4a, 0, &COPYDATA, , ahk_id %HWND% ;SendMessage waits for the target window to process the message, up until the timeout period expires.
}

global XY_eval_and_returnvalue_XYRETURN

XY_eval_and_returnvalue(expression)          ;Send message to XYplorer
{
  global XY_eval_and_returnvalue_XYRETURN
  SenderHWND := A_ScriptHwnd + 0  ;Return this script's hidden hwdn id.  +0 to convert from Hex to Dec
  MessagetoXYplorer := "::CopyData " SenderHWND ", " expression ", 0"    ;resolved to sth like this:   ::CopyData 7409230, <curitem>, 0   _OR like this:  ::CopyData 7409230, tab('get','count'), 0
  SetTitleMatchMode, 2
  HWND := WinExist("XYplorer ahk_class ThunderRT6FormDC")
  Size := StrLen(MessagetoXYplorer)
  if !(A_IsUnicode)
  {
  VarSetCapacity(Data, Size * 2, 0)
  StrPut(MessagetoXYplorer, &Data, Size, "UTF-16")
  }
  Else
  Data := MessagetoXYplorer

  VarSetCapacity(COPYDATA, A_PtrSize * 3, 0)
  NumPut(4194305, COPYDATA, 0, "Ptr")
  NumPut(Size * 2, COPYDATA, A_PtrSize, "UInt")
  NumPut(&Data, COPYDATA, A_PtrSize * 2, "Ptr")
  OnMessage(0x4a, "Function_Receive_WM_COPYDATA")  ; 0x4a is WM_COPYDATA. This onhold and wait for the WM_Copydata from XYplorer then execute Function_Receive_WM_COPYDATA(wParam, lParam) below
  SendMessage, 0x4a, 0, &COPYDATA, , ahk_id %HWND% ;SendMessage waits for the target window to process the message, up until the timeout period expires.
  OnMessage(0x4a, "")
  return XY_eval_and_returnvalue_XYRETURN
}

Function_Receive_WM_COPYDATA(wParam, lParam)
{
  global XY_eval_and_returnvalue_XYRETURN
  StringAddress := NumGet(lParam + 2*A_PtrSize) ;lParam+8 is the address of CopyDataStruct's lpData member.
  CopyOfData := StrGet(StringAddress)    ;May also specify CP0 (default) or UTF-8 or UTF-16:   StrGet(StringAddress, NumGet(lParam+A_PtrSize), "UTF-16")
  cbData := NumGet(lParam+A_PtrSize)/2  ;cbData/2 = String length
  StringLeft, Datareceived, CopyOfData, cbData
  XY_eval_and_returnvalue_XYRETURN := Datareceived
}

;*****************************************************************************************************************
